import logging
import adage
import adage.node
import adage.serialize
import jsonpointer
import time
import uuid
import jsonpath_rw
import yadagestep
log = logging.getLogger(__name__)

class stage_base(object):
    '''
    Base class for workflow stages (i.e. extension rules)
    provides common datastructures and the extension predicate.
    Implementations are required to provide a schedule()
    method that is called upon apply
    '''
    def __init__(self,name,context,dependencies = None):
        self.view = None
        self.name = name
        self.context = context
        self.depspec = dependencies

    def applicable(self,flowview):
        if not self.depspec: return True
        from handlers.predicate_handlers import handlers as pred_handlers
        predicate = pred_handlers[self.depspec['dependency_type']]
        return predicate(flowview,self.depspec)

    def apply(self,flowview):
        self.view = flowview
        self.schedule()

    def addStep(self,step):
        dependencies = [self.view.dag.getNode(k.stepid) for k in step.inputs]
        self.view.addStep(step, stage = self.name, depends_on = dependencies)

    def addWorkflow(self,rules, initstep):
        self.view.addWorkflow(rules, initstep = initstep, stage = self.name)

    #(de-)serialization
    def json(self):
        return {
            'name':self.name,
            'context':self.context,
            'dependencies':self.depspec
        }

class initStage(stage_base):
    '''
    simple stage that just adds a initializer step to the DAG
    '''
    def __init__(self, step, context, dependencies):
        super(initStage,self).__init__('init', context,dependencies)
        self.step = step

    def schedule(self):
        log.debug('initializing a scope with init step: %s',self.step.prepublished)
        self.addStep(self.step)

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        instance = cls(
            step = yadagestep.initstep.fromJSON(data['step']),
            context = data['context'],
            dependencies = data['dependencies']
        )
        return instance

    def json(self):
        data = super(initStage,self).json()
        data.update(type = 'initStage', info = '', step = self.step.json())
        return data

class jsonStage(stage_base):
    '''
    A stage that is defined via the JSON scheduler schemas
    '''
    def __init__(self,json,context):
        self.stageinfo = json['scheduler']
        super(jsonStage,self).__init__(json['name'],context,json['dependencies'])

    def schedule(self):
        from yadage.handlers.scheduler_handlers import handlers as sched_handlers
        scheduler = sched_handlers[self.stageinfo['scheduler_type']]
        scheduler(self,self.stageinfo)

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        return cls(json = {
            'scheduler':data['info'],
            'name':data['name'],
            'dependencies':data['dependencies']
        }, context = data['context'])

    def json(self):
        data = super(jsonStage,self).json()
        data.update(type = 'jsonStage', info = self.stageinfo)
        return data

class offsetRule(object):
    '''
    A wrapper object around a scoped rule, so that it can be applied from
    a global p.o.v., i.e. as adage expects its rules.
    '''
    def __init__(self,rule,offset = None, identifier = None):
        self.identifier = identifier or str(uuid.uuid4())
        self.rule = rule
        self.offset = offset

    def applicable(self,adageobj):
        return self.rule.applicable(WorkflowView(adageobj,self.offset))

    def apply(self,adageobj):
        self.rule.apply(WorkflowView(adageobj,self.offset))

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        if data['rule']['type'] == 'initStage':
            rule = initStage.fromJSON(data['rule'])
        elif data['rule']['type'] == 'jsonStage':
            rule = jsonStage.fromJSON(data['rule'])
        return cls(
                    rule = rule,
                    identifier = data['id'],
                    offset = data['offset']
                   )

    def json(self):
        return {'type':'offset',
                'id':self.identifier,
                'offset':self.offset,
                'rule':self.rule.json()}

class YadageNode(adage.node.Node):
    '''
    Node object for yadage that extends the default with
    the ability to have prepublished results
    '''
    def __init__(self,name,task,identifier = None):
        super(YadageNode,self).__init__(name,task,identifier)

    def __repr__(self):
        lifetime = time.time()-self.define_time
        return '<YadageNode {} {} lifetime: {} (id: {}) has result: {}>'.format(self.name,self.state,lifetime,self.identifier,self.has_result())

    def has_result(self):
        return (self.task.prepublished is not None) or self.successful()

    @property
    def result(self):
        if self.task.prepublished is not None:
            return self.task.prepublished
        return super(YadageNode,self).result

    @classmethod
    def fromJSON(cls,data):
        if data['task']['type'] == 'initstep':
            task = yadagestep.initstep.fromJSON(data['task'])
        elif data['task']['type'] == 'yadagestep':
            task = yadagestep.yadagestep.fromJSON(data['task'])
        return cls(data['name'],task,data['id'])


def json_or_nil(x):
    return None if x is None else x.json()

class YadageWorkflow(adage.adageobject):
    '''
    The overall workflow state object that extends the basic
    Adage state object by two bookkeeping structures.
    '''
    def __init__(self):
        super(YadageWorkflow,self).__init__()
        self.stepsbystage = {}
        self.bookkeeping = {}

    def view(self,offset = ''):
        return WorkflowView(self,offset)

    def json(self):
        from adage.serialize import obj_to_json
        data = obj_to_json(self,
                           ruleserializer  = json_or_nil,
                           taskserializer  = json_or_nil,
                           proxyserializer = json_or_nil,
                           )
        data['bookkeeping']  = self.bookkeeping
        data['stepsbystage'] = self.stepsbystage
        return data

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data,proxyclass,backend = None):
        rules, applied = [], []
        for x in data['rules']:
            rules += [offsetRule.fromJSON(x)]
        for x in data['applied']:
            applied += [offsetRule.fromJSON(x)]

        instance = cls()
        instance.rules = rules
        instance.applied_rules = applied
        instance.bookkeeping = data['bookkeeping']
        instance.stepsbystage = data['stepsbystage']
        instance.dag = adage.serialize.dag_from_json(
            data['dag'],
            YadageNode,
            proxyclass,
            backend
        )
        return instance

    @classmethod
    def createFromJSON(cls,jsondata,context):
        instance = cls()
        rules = [jsonStage(yml,context) for yml in jsondata['stages']]
        rootview = WorkflowView(instance)
        rootview.addWorkflow(rules)
        return instance

def createOffsetMeta(offset,bookkeeping):
    pointer = jsonpointer.JsonPointer(offset)
    parts = pointer.parts
    view = bookkeeping
    for x in parts:
        if not x in view: view[x] = {}
        view = view[x]
    scoped = pointer.resolve(bookkeeping)
    if not '_meta' in scoped:
        scoped['_meta'] = {'rules':[],'steps':[]}

class WorkflowView(object):
    '''
    Provides a 'view' of the overall workflow object that corresponds
    to a particular level of nesting. That is, this presents the scope
    within which extension rules operate (i.e. add steps, reference
    other steps, etc)
    '''
    def __init__(self,workflowobj,offset = ''):
        self.dag           = workflowobj.dag
        self.rules         = workflowobj.rules
        self.applied_rules = workflowobj.applied_rules
        self.offset        = offset
        self.steps         = jsonpointer.JsonPointer(self.offset).resolve(workflowobj.stepsbystage)
        self.bookkeeper    = jsonpointer.JsonPointer(self.offset).resolve(workflowobj.bookkeeping)

    def query(self,query,collection):
        matches = jsonpath_rw.parse(query).find(collection)
        return matches

    def getSteps(self,query):
        result =  [self.dag.getNode(step['_nodeid']) for match in self.query(query,self.steps) for step in match.value]
        return result

    def init(self, initdata, name = 'init'):
        step = yadagestep.initstep(name,initdata)
        self.addRule(initStage(step,{},None),self.offset)

    def addRule(self,rule,offset = ''):
        thisoffset = jsonpointer.JsonPointer(offset)
        if self.offset:
            fulloffset = jsonpointer.JsonPointer.from_parts(jsonpointer.JsonPointer(self.offset).parts + thisoffset.parts).path
        else:
            fulloffset = thisoffset.path
        offsetrule = offsetRule(rule,fulloffset)
        self.rules += [offsetrule]
        createOffsetMeta(thisoffset.path,self.bookkeeper)
        thisoffset.resolve(self.bookkeeper)['_meta']['rules'] += [offsetrule.identifier]
        return offsetrule.identifier

    def addStep(self,step, stage, depends_on = None):
        node = YadageNode(step.name,step)
        self.dag.addNode(node,depends_on = depends_on)

        log.debug('added node %s',node)

        noderef = {'_nodeid':node.identifier}
        if stage in self.steps:
            self.steps[stage] += [noderef]
        else:
            self.steps[stage]  = [noderef]
        self.bookkeeper['_meta']['steps'] += [node.identifier]

    def addWorkflow(self, rules, initstep = None, stage = None):
        if initstep:
            rules += [initStage(initstep,{},None)]
        newsteps = {}
        if stage in self.steps:
            self.steps[stage] += [newsteps]
        elif stage is not None:
            self.steps[stage]  = [newsteps]

        offset = jsonpointer.JsonPointer.from_parts([stage,len(self.steps[stage])-1]).path if stage else ''
        if stage is not None:
            self.steps[stage][-1]['_offset'] = offset

        for rule in rules:
            self.addRule(rule,offset)



import jsonpointer
import shutil
import os
import networkx as nx

def select_rule(workflow,offset,name):
    for x in workflow.applied_rules:
        if x.offset == offset and x.rule.name == name:
            return x
    raise RuntimeError('rule not found')

def stepsofrule(workflow,offset,name):
    rule = select_rule(workflow,offset,name)
    path = '/'.join([rule.offset,rule.rule.name])
    p = jsonpointer.JsonPointer(path)
    return rule,[x['_nodeid'] for x in p.resolve(workflow.stepsbystage)]

def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None

def reset_step(workflow,step):
    s = workflow.dag.getNode(step)
    reset_node_state(s)
    try:
        for rw in s.task.context['readwrite']:
            shutil.rmtree(rw)
            os.makedirs(rw)
    except AttributeError:
        pass

def reset_steps(workflow,steps):
    for s in steps: reset_step(workflow,s)

def collective_downstream(workflow,steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag,step):
            downstream.add(x)
    return list(downstream)

def reset_state(workflow,offset,name):
    rule, steps = stepsofrule(workflow,offset,name)
    toreset = steps + collective_downstream(workflow,steps)
    reset_steps(workflow,toreset)



import logging
import utils
import jsonpointer
import jsonpath_rw

log = logging.getLogger(__name__)

handlers,scheduler = utils.handler_decorator()

def convert(thing):
    if type(thing)==jsonpath_rw.jsonpath.Index:
        return thing.index
    if type(thing)==jsonpath_rw.jsonpath.Fields:
        fs = thing.fields
        assert len(fs)==1
        return fs[0]

def unravelpath(path):
    if type(path)==jsonpath_rw.jsonpath.Child:
        for x in unravelpath(path.left):
            yield x
        yield convert(path.right)
    else:
        yield convert(path)

def path2pointer(path):
    return jsonpointer.JsonPointer.from_parts(x for x in unravelpath(path)).path

def checkmeta(flowview,metainfo):
    log.debug('checking meta %s on view with offset %s',metainfo,flowview.offset)
    applied_ids = [rl.identifier for rl in flowview.applied_rules]
    rulesok = all([x in applied_ids for x in metainfo['rules']])

    # for x in metainfo['steps']:
    #     print 'checking step',x,flowview.dag.getNode(x).has_result
    stepsok = all([flowview.dag.getNode(x).has_result() for x in metainfo['steps']])
    log.debug('all rules applied: %s, all steps have results: %s',rulesok,stepsok)
    return (rulesok and stepsok)


def scope_done(scope,flowview):
    log.debug('checking scope %s on view with offset %s',scope,flowview.offset)
    result = True

    bookkeeper = jsonpointer.JsonPointer(scope).resolve(flowview.bookkeeper)
    for k,v in bookkeeper.iteritems():
        for k,v in bookkeeper.iteritems():
            if k=='_meta':
                result = result and checkmeta(flowview,v)
            else:
                childscope = scope+'/{}'.format(k)
                result = result and scope_done(childscope,flowview)
    return result

@scheduler('jsonpath_ready')
def jsonpath_ready(flowview,depspec):
    log.debug('checking jsonpath ready predicate\n%s',depspec)
    dependencies = depspec['expressions']
    for x in dependencies:
        depmatches = flowview.query(x,flowview.steps)
        if not depmatches:
            return False
        issubwork = '_nodeid' not in depmatches[0].value[0]
        if issubwork:
            log.debug('dependency is a subworkflow. determine if scope is done')
            if not all([scope_done(scope['_offset'],flowview) for match in depmatches for scope in match.value]):
                return False
        else:
            if not all([x.has_result() for x in flowview.getSteps(x)]):
                return False
    log.debug('all checks ok, predicate is True')
    return True



import logging
import utils
import jsonpointer
import itertools
import os
import copy
import jq
import jsonpath_rw

from yadage.yadagestep import yadagestep, initstep, outputReference
from yadage.yadagemodels import jsonStage

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step attributes are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])


def pointerize(jsondata, asref=False, stepid=None):
    '''
    a helper method that replaces leaf nodes in a JSON object with
    a outputReference objects (~ a JSONPath) pointing to that leaf position
    useful to track access to leaf nodes later on.
    '''
    allleafs = jq.jq('leaf_paths').transform(jsondata, multiple_output=True)
    leafpointers = [jsonpointer.JsonPointer.from_parts(x).path for x in allleafs]
    jsondata_proxy = copy.deepcopy(jsondata)
    for leaf in leafpointers:
        x = jsonpointer.JsonPointer(leaf)
        x.set(jsondata_proxy, outputReference(stepid, x) if asref else x.path)
    return jsondata_proxy


def select_reference(step, selection):
    '''
    resolves a jsonpath selection and returns JSONPointerized matches
    '''
    log.debug('selecting output from step %s',step)
    pointerized = pointerize(step.result, asref=True, stepid=step.identifier)
    matches = jsonpath_rw.parse(selection).find(pointerized)
    if not matches:
        log.error('no matches found for selection %s in result %s', selection, step.result)
        raise RuntimeError('no matches found in reference selection')

    if len(matches) > 1:
        log.error('found multiple matches to query: %s within result: %s\n \ matches %s', selection, step.result, matches)
        raise RuntimeError('multiple matches in result jsonpath query')
    return matches[0].value


def combine_outputs(outputs, flatten, unwrapsingle):
    combined = []
    for reference in outputs:
        if type(reference)==list:
            if flatten:
                for elementref in reference:
                    combined+=[elementref]
            else:
                combined+=[reference]
        else:
            combined+=[reference]
    if len(combined)==1 and unwrapsingle:
        combined = combined[0]
    return combined

def select_steps(stage,query):
    return stage.view.getSteps(query)


def select_outputs(steps,selection,flatten,unwrapsingle):
    return combine_outputs(map(lambda s: select_reference(s, selection), steps), flatten, unwrapsingle)


def resolve_reference(stage,selection):
    '''resolves a output reference by selecting the stage and stage outputs'''
    log.debug('resolving selection %s',selection)
    if type(selection) is not dict:
        return None
    else:
        steps = select_steps(stage, selection['stages'])
        log.debug('selected steps %s',steps)
        outputs = select_outputs(steps,
                                 selection['output'],
                                 selection.get('flatten', False),
                                 selection.get('unwrap', False))
        log.debug('selected outputs %s',outputs)
        return outputs


def select_parameter(stage,parameter):
    if type(parameter) is not dict:
        value = parameter
    else:
        value = resolve_reference(stage,parameter)
    return value


def finalize_value(stage,step,value,context):
    '''
    finalize a value by recursively resolving references and
    interpolating with the context when necessary
    '''
    if type(value)==outputReference:
        step.used_input(value)
        v = value.pointer.resolve(stage.view.dag.getNode(value.stepid).result)
        return finalize_value(stage,step,v,context)
    if type(value)==list:
        return [finalize_value(stage,step,x,context) for x in value]
    if type(value) in [str,unicode]:
        return value.format(**context)
    return value

def finalize_input(stage,step,json,context):
    '''
    evaluate final values of parameters by either resolving a
    reference to a upstream output or evaluating a static
    reference from the template (possibly string-interpolated)
    '''
    context = context.copy()
    context['workdir'] = context['readwrite'][0]
    result = {}
    for k,v in json.iteritems():
        if type(v) is not list:
            result[k] = finalize_value(stage,step,v,context)
        else:
            result[k] = [finalize_value(stage,step,element,context) for element in v]
    return result

def step_or_init(name,spec,context):
    if 'step' in spec:
        stepcontext = make_new_context(name,context)
        return yadagestep(name = name, spec = spec['step'], context = stepcontext)
    elif 'workflow' in spec:
        return initstep('init {}'.format(name))

def make_new_context(name,oldcontext):
    '''
    creates a new context from an existing context.
    specifically it declares a new read-write nested under the old
    context's read-write and adds all read-write and read-only locations
    of the old context as read-only
    '''
    newcontext = {
        'nametag':name,
        'readwrite':['{}/{}'.format(oldcontext['readwrite'][0],name)],
        'readonly':[]
    }
    newcontext['readonly'] += [ro for ro in itertools.chain(oldcontext['readonly'],oldcontext['readwrite'])]
    os.makedirs(newcontext['readwrite'][0])
    return newcontext

def addStepOrWorkflow(name,stage,step,spec):
    if type(step)==initstep:
        newcontext = make_new_context(name,stage.context)
        subrules = [jsonStage(yml,newcontext) for yml in spec['workflow']['stages']]
        stage.addWorkflow(subrules, initstep = step)
    else:
        stage.addStep(step)

def get_parameters(spec):
    return {x['key']:x['value']for x in spec['parameters']}

@scheduler('singlestep-stage')
def singlestep_stage(stage,spec):
    '''
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs
    '''
    log.debug('scheduling singlestep stage with spec:\n%s',spec)


    parameters = {
        k:select_parameter(stage,v) for k,v in get_parameters(spec).iteritems()
    }

    step = step_or_init(name = stage.name, spec = spec, context = stage.context)

    ctx = step.context if hasattr(step,'context') else stage.context
    finalized = finalize_input(stage,step,parameters,ctx)

    addStepOrWorkflow(stage.name,stage,step.s(**finalized),spec)

def scatter(parameters,scatter):
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter['parameters']:
        to_scatter[scatpar] = commonpars.pop(scatpar)

    singlesteppars=[]
    if scatter['method']=='zip':
        keys, zippable = zip(*[(k,v) for i,(k,v) in enumerate(to_scatter.iteritems())])

        for zipped in zip(*zippable):
            individualpars = dict(zip(keys,zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter['method']=='cartesian':
        for what in itertools.product(*[to_scatter[k] for k in scatter['parameters']]):
            individualpars = dict(zip(scatter['keys'],what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars

@scheduler('multistep-stage')
def multistep_stage(stage,spec):
    '''
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported
    'zip': one or more arrays of length n are iterated through in lock-step.
           n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    'cartesian': a cartesian product of a number of arrays (possibly different sizes)
                 adds n1 x n2 x ... nj nodes.
    Nodes are attached to the DAG based on used upstream inputs
    '''
    log.debug('scheduling multistep stage with spec:\n%s',spec)
    parameters = {
        k:select_parameter(stage,v) for k,v in get_parameters(spec).iteritems()
    }
    singlesteppars = scatter(parameters,spec['scatter'])

    for i,pars in enumerate(singlesteppars):
        index_context = stage.context.copy()
        index_context.update(index=i)

        singlename = '{}_{}'.format(stage.name,i)
        step = step_or_init(name = singlename, spec = spec, context = stage.context)
        finalized = finalize_input(stage,step,pars,index_context)

        addStepOrWorkflow(singlename,stage,step.s(**finalized),spec)



#!/usr/bin/env python
import os
import click
import jsonschema
import logging
import workflow_loader
import capschemas
import json
import jsonref

class ref(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, jsonref.JsonRef):
            return obj.copy()
        return json.JSONEncoder.default(self, obj)


logging.basicConfig(level = logging.ERROR)
log = logging.getLogger(__name__)
@click.command()
@click.argument('workflow')
@click.argument('toplevel', default = '')
@click.argument('schemadir', default = '')
@click.option('--stdout','-s', default = False, is_flag=True)
def main(workflow,toplevel,schemadir,stdout):
    if not toplevel:
        toplevel = os.getcwd()
    if not schemadir:
        schemadir = capschemas.schemadir
    try:
        data = workflow_loader.workflow(workflow, toplevel = toplevel, schemadir = schemadir, validate = True)
        if stdout:
            print json.dumps(data,cls = ref)
        else:
            click.secho('workflow validates against schema', fg = 'green')
    except jsonschema.exceptions.ValidationError:
        log.exception('validation error')
        click.secho('workflow does not validate against schema', fg = 'red')
    except:
        log.exception('')
        click.secho('this is not even wrong (non-ValidationError exception)', fg = 'red')

        
if __name__ == '__main__':
    main()



import packtivity
import logging
import jsonpointer
log = logging.getLogger(__name__)

class outputReference(object):
    def __init__(self,stepid,pointer):
        self.stepid = stepid
        self.pointer = pointer

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        return cls(data['stepid'],jsonpointer.JsonPointer(data['pointer_path']))

    def json(self):
        return {
            'stepid':self.stepid,
            'pointer_path':self.pointer.path
        }

class stepbase(object):
    def __init__(self,name):
        self.name = name
        self.inputs = []
        self.attributes = {}
        self.prepublished = None

    def used_input(self,reference):
        self.inputs += [reference]

    #(de-)serialization
    def json(self):
        return {
            'name':self.name,
            'attributes':self.attributes,
            'prepublished':self.prepublished,
            'inputs':[x.json() for x in self.inputs]
        }

class initstep(stepbase):
    def __init__(self,name, initdata = None):
        super(initstep,self).__init__(name)
        self.prepublished = None
        if initdata is not None:
            self.s(**initdata)

    def __call__(self):
        pass

    def s(self,**attributes):
        self.attributes = attributes
        self.prepublished = self.attributes
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        instance =  cls(data['name'])
        instance.attributes = data['attributes']
        instance.prepublished = data['prepublished']
        instance.inputs = map(outputReference.fromJSON,data['inputs'])
        return instance

    def json(self):
        data = super(initstep,self).json()
        data.update(type = 'initstep')
        return data

class yadagestep(stepbase):
    def __init__(self,name,spec,context):
        super(yadagestep,self).__init__(name)
        self.spec = spec
        self.context = context

    def s(self,**attributes):
        self.attributes.update(**attributes)
        #attempt to prepublish output data merely from inputs
        #will still be None if not possible
        self.prepublished = packtivity.prepublish(self.spec,self.attributes,self.context)
        log.debug('parameters for yadagestep set to %s. prepublished result, if any: %s',
            self.attributes,
            self.prepublished
        )
        return self

    #(de-)serialization
    @classmethod
    def fromJSON(cls,data):
        instance =  cls(data['name'],data['spec'],data['context'])
        instance.attributes = data['attributes']
        instance.prepublished = data['prepublished']
        instance.inputs = map(outputReference.fromJSON,data['inputs'])
        return instance

    def json(self):
        data = super(yadagestep,self).json()
        data.update(
            type = 'yadagestep',
            spec = self.spec,
            context = self.context,
        )
        return data



from setuptools import setup, find_packages

setup(
  name = 'yadage',
  version = '0.3.0',
  description = 'yadage - YAML based adage',
  url = '',
  author = 'Lukas Heinrich',
  author_email = 'lukas.heinrich@cern.ch',
  packages = find_packages(),
  include_package_data = True,
  install_requires = [
    'functools32',
    'adage>0.3.0',
    'packtivity>0.0.1',
    'cap-schemas',
    'click',
    'psutil',
    'requests[security]>=2.9',
    'pyyaml',
    'jsonref',
    'jsonschema',
    'jsonpointer>=1.10',
    'jsonpath_rw',
    'packtivity',
    'adage',
    'jq'
  ],
  extras_require = {
    'celery' : ['celery','redis']
  },
  entry_points = {
      'console_scripts': [
          'yadage-run=yadage.steering:main',
          'yadage-manual=yadage.manualcli:mancli',
          'yadage-validate=yadage.validator_workflow:main',
      ],
  },
  dependency_links = [
  ]
)



import yadage.yadagestep
import adage.backends
from celery import shared_task
from celery.result import AsyncResult
from packtivity import packtivity
from packtivitybackend import AdagePacktivityBackendBase, PacktivityProxyBase

@shared_task
def packtivity_task(spec, attributes, context):
    '''celery enabled packtiity'''
    print 'running a packtivity via a celery task'
    return packtivity(spec,attributes,context)

@shared_task
def init_task():
    '''dummy task'''
    print 'running a dummytask used for init steps'

class PacktivityCeleryProxy(PacktivityProxyBase):
    def proxyname(self):
        return 'PacktivityCeleryProxy'

    def details(self):
        return {
            'task_id':self.proxy.task_id,
            'task_name':self.proxy.task_name
        }

    @classmethod
    def fromJSON(cls,data):
        proxy = AsyncResult(
            data['proxydetails']['task_id'],
            task_name = data['proxydetails']['task_name']
        )
        return cls(proxy)


class PacktivityCeleryBackend(AdagePacktivityBackendBase):
    def __init__(self,app):
        super(PacktivityCeleryBackend,self).__init__(adage.backends.CeleryBackend(app))

    def submit(self,task):
        '''
        if the task type is a genuine yadagestep we submit is as a packtivity callable,
        if it's an init step, which has a trivial call body, we'll pass a dummy task
        '''
        tasktype = type(task)
        if tasktype == yadage.yadagestep.yadagestep:
            resultproxy = packtivity_task.delay(task.spec,task.attributes,task.context)
        elif tasktype == yadage.yadagestep.initstep:
            resultproxy = init_task.delay()
        else:
            raise RuntimeError('cannot figure out how to submit a task of type {}'.format(tasktype))

        return PacktivityCeleryProxy(resultproxy)
